#!/usr/bin/env python3
"""Examples to show xtp_binds usage."""
import io
from contextlib import redirect_stdout


def capture_standard_output(function, *args, **kwargs):
    """Capture standard output of a given function."""
    with redirect_stdout(io.StringIO()) as stdout:
        function(*args, **kwargs)

    return stdout.getvalue()



"""Module to edit the options in the XML file."""

import xml.etree.ElementTree as ET
import os
from lxml import etree
from lxml.etree import Element, ElementTree
from copy import deepcopy

def update_xml_text(xml: ElementTree, dict_options: dict) -> ElementTree:
    """update default option files with the user define options

    Args:
        xml_tree (ElementTree): tree of the xml file
        dict_options (dict): dictionary of user defined options

    Returns:
        ET.ElementTree: xml structure containing the update options.
    """

    # copy the xml tree
    xml_cpy = deepcopy(xml)

    # replace text value of the dict elements
    for key, value in dict_options.items():
        try:
            xml_cpy.find('.//'+key).text = str(value)
        except Exception as e:
            print(e)
            print(key, ' not found ')
    return xml_cpy


def insert_linked_xmlfiles(tree: ElementTree, base_path: str) -> None:
    """insert the tree of linked xml files in the original tree

    Args:
        tree (ElementTree): original xml tree containing linked files
        base_path (str): base path for the linked files
    """

    def _recursive_insert_links(el: Element):
        """recursively add links files to the main tree

        Args:
            el (Element): lxml element

        Raises:
            ValueError: If one of the link is not found
        """

        for child in el.getchildren():
            _recursive_insert_links(child)

        if 'link' in el.attrib:
            link_name = os.path.join('subpackages', el.attrib['link'])
            link_name = os.path.join(base_path, link_name)
            if not os.path.isfile(link_name):
                raise ValueError('File %s not found' %link_name)
            xml_link = etree.parse(link_name)
            el.getparent().replace(el, xml_link.getroot())

    # recursively insert the links
    _recursive_insert_links(tree.getroot())

    return tree


def remove_empty_text_elements(tree: ET.ElementTree, remove_attributes=[]) -> ElementTree:
    """Recursively remove all empty nodes and elements containing only empty nodes

    Args:
        xml_tree (ElementTree): tree of the xml file
        remove_attributes (list): list of xml attributes to remove
    """

    def is_empty(elem: Element) -> bool:
        """returns true if elem does not have text

        Args:
            elem (Element): xml element

        Returns:
            bool true if e is empty
        """
        return (elem.text is None) or (elem.text.strip() == '')

    def recursively_remove_empty_elements(elem: Element):
        """reccursive function that remove empty elements

        Args:
            e (Element): xml element

        Returns:
            bool: True if we need to remove the element False otherwise
        """

        # recursive call that gets a list of :
        # [] if e doens't have children
        # [True, False, False, True ] where each bool indicate if e children are empty or not
        child_bool = list(recursively_remove_empty_elements(c) for c in elem.iterchildren())
        nchildren = len(child_bool)

        # there is an issue with the last element
        # so we need to catch the exception
        try:

            # if e doesn't have kids, i.e. it's a leaf
            # remove e if it doesn't have text and returns True
            # keep e if it has text and return False
            if nchildren == 0:
                if is_empty(elem):
                    elem.getparent().remove(elem)
                    return True
                else:
                    return False
            # remove e if all its children have no text
            # and returns True
            elif all(child_bool):
                elem.getparent().remove(elem)
                return True

        # excpetion for root ...
        except:
            return False


    # remove attributes if any are passed
    etree.strip_attributes(tree, remove_attributes)

    # remove the empties
    recursively_remove_empty_elements(tree.getroot())

    return tree


"""Utilities and constants."""
import scipy.constants
import numpy as np

H2EV = scipy.constants.physical_constants['Hartree energy in eV'][0]
BOHR2ANG = scipy.constants.physical_constants['Bohr radius'][0] * 1.e10
INVCM2EV = (scipy.constants.c *
            scipy.constants.physical_constants['Planck constant in eV/Hz'][0])*100.0
AFU2INVS = np.sqrt((scipy.constants.physical_constants['Hartree energy'][0]/np.power(
    scipy.constants.physical_constants['Bohr radius'][0], 2))/scipy.constants.physical_constants['atomic mass constant'][0])
AFU2INVCM = AFU2INVS/(2.0*np.pi*scipy.constants.c*100.)
AFU2EV = scipy.constants.physical_constants['reduced Planck constant in eV s'][0] * AFU2INVS



import unittest
import pytest
from ase.build import molecule
import os
try:
    from pyxtp import xtp
    from pyxtp.options import Options
except ImportError:
    raise unittest.SkipTest('pyxtp not available')


@pytest.fixture(scope="module")
def dftgwbse_xml():
    votcashare = os.environ.get("VOTCASHARE")
    if votcashare is None:
        msg = (
            "pyxtp: cannot find Votca installation, "
            "please set the VOTCASHARE environment variable"
        )
        raise RuntimeError(msg)
    return f"{votcashare}/xtp/xml/dftgwbse.xml"


class TestXTP:

    @pytest.fixture
    def CO(self):
        calculator = xtp(nthreads=2)
        atoms = molecule('CO', positions=([0,0,0],[1.4,0,0]), calculator=calculator)
        return atoms

    @pytest.mark.xfail()
    def test_failed_option(self, CO):
        CO.calc.options.dftpackage.basisset = 'invalid_basis_set'
        CO.get_potential_energy()

    def test_set(sefl, CO):
        CO.calc.set(basis='def2-svp', xc='lda', charge=1)
        if CO.calc.options.dftpackage.basisset != 'def2-svp':
            raise AssertionError("Error in test_set")
        if CO.calc.options.dftpackage.functional != 'lda':
            raise AssertionError("Error in test_set")
        if CO.calc.options.dftpackage.charge != 1:
            raise AssertionError("Error in test_set")

    def test_set_default(self, CO):
        CO.calc.set(basis='sto-3g')
        CO.calc.set_default_options()
        if CO.calc.options.dftpackage.basisset != '':
            raise AssertionError("Error in test_set_default")

    def test_set_from_options(self, CO, dftgwbse_xml):
        opt = Options(dftgwbse_xml)
        opt.dftpackage.basisset = 'sto-3g'
        CO.calc.set_from_options(opt)
        if CO.calc.options.dftpackage.basisset != 'sto-3g':
            raise AssertionError("Error in test_set_from_options")

    @pytest.mark.xfail()
    def test_failed_get_total_energy(self, CO):
        CO.get_total_energy(name='invalid_energy_name')

    def test_get_total_energy(self, CO):
        CO.get_potential_energy()

        new_atoms = xtp.read_atoms('CO.orb')
        if(CO.calc.results['energy'] != new_atoms.calc.results['energy']):
            raise AssertionError("Error in test_get_total_energy")

        for name in CO.calc.implemented_properties:
            if name not in ["forces", "oscillator_strength"]:
                if CO.calc.get_total_energy(name, level=1, dynamic = False) == None:
                    raise AssertionError("Error in test_get_total_energy")






#!/usr/bin/env python3
#
# Copyright 2009-2020 The VOTCA Development Team (http://www.votca.org)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Module to extract metadata from the xml files."""

import argparse
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Iterable, List, Optional, Tuple

msg = "extract_metadata.py -i file.xml"

parser = argparse.ArgumentParser(description=msg)
parser.add_argument('-i', required=True,
                    help="Input file in XML format")
parser.add_argument('-o', help="Optional output file", default=None)
parser.add_argument("-m", "--mode", help="Operation mode: xtp, csg",
                    choices=["xtp", "csg"], default="xtp")

MAXIMUM_LINE_LENGTH = 60

DEFAULTS_TABLE_HEADER = """
.. list-table::
   :header-rows: 1
   :widths: 30 20 15 15
   :align: center

   * - Property Name
     - Default Value
     - Description
     - Valid Input"""


def xtp_table_header(x: str) -> str:
    """Create a table for XTP calculators."""
    return f"""
{x}
The following table contains the defaults input options for the calculator, The default `OPTIONAL` means this option is switched off, if no input is given. 
`REQUIRED` arguments have to be specified, otherwise an error is thrown.
{DEFAULTS_TABLE_HEADER}"""


def csg_table_header(x: str):
    """Create a CSG table with its default values."""
    return f"""
{x}
The following table contains the input options for CSG,

.. list-table::
   :header-rows: 1
   :align: center

   * - Property Name
     - Default Value
     - Description"""


XTP_TABLE_LINE = """
   * - {}
     - {}
     - {}
     - {}""".format

CGS_TABLE_LINE = """
   * - {}
     - {}
     - {}""".format


def main():
    """Parse the command line arguments and run workflow."""
    args = parser.parse_args()
    file_name = Path(args.i)
    if args.mode == "xtp":
        table = xtp_create_rst_table(file_name)
    elif args.mode == "csg":
        table = csg_create_rst_table(file_name)

    # Print
    if args.o is not None:
        path = Path(args.o)
        create_parent_folder(path)
        with open(path, 'w')as f:
            f.write(table)
    else:
        print(table)


def get_root_children(file_name: Path) -> List[ET.Element]:
    """Get all the node children from root."""
    tree = ET.parse(file_name)
    root = tree.getroot()
    resolve_links(file_name, root)
    return list(root)


def resolve_links(file_name: Path, elem: ET.Element):
    if "link" in elem.attrib:
        s=elem.attrib.get("link")
        #python split method cannot split for "," and " " so we replace all "," by " "
        s.replace(',', ' ')
        links = elem.attrib.get("link").split(" ")
        for link in links:
            package_path = file_name.parent/"subpackages"/link
            tree = ET.parse(package_path)
            root = tree.getroot()
            for child in root:
                elem.append(child)
    for child in elem:
        resolve_links(file_name, child)


def xtp_extract_metadata(file_name: Path) -> Tuple[str, ET.Element]:
    """Get the description and elements from the xml file."""
    data = get_root_children(file_name)[0]
    header = data.attrib.get("help", "")
    return header, data


def csg_extract_metadata(file_name: Path) -> Tuple[str, List[ET.Element]]:
    """Get the description and elements from the xml file."""
    children = get_root_children(file_name)
    header = children[0].text
    return header, children[1:]


def xtp_get_recursive_attributes(elem: ET.Element, root_name: str = "") -> str:
    """Get recursively the attributes of an ``ET.Element``."""
    s = ""
    name = root_name + elem.tag
    if list(elem):
        return ''.join(xtp_get_recursive_attributes(el, f"{name}.") for el in list(elem))

    description = split_line(elem.attrib.get("help", ""))
    default = split_line(elem.attrib.get("default", ""))
    choices = multiline(elem.attrib.get("choices", ""))
    s += XTP_TABLE_LINE(name, default, description, choices)

    return s


def csg_get_recursive_attributes(elem: ET.Element, root_name: str = "") -> str:
    """Get recursively the attributes of an ``ET.Element``."""
    s = ""
    if list(elem):
        children = iter(elem)
        desc_elem = elem.find("DESC")
        if desc_elem is not None:
            description = desc_elem.text
        else:
            description = ""
        description = split_line(description)

        name = root_name + elem.tag
        default = "" if elem.text is None else ' '.join(elem.text.split())
        s += CGS_TABLE_LINE(name, default, description)
        s += ''.join(
            csg_get_recursive_attributes(
                el, f"{name}.") for el in children if el.tag != "DESC")
        return s

    return s


def xtp_create_rst_table(file_name: Path) -> str:
    """Create an RST table using the metadata in the XML file."""
    header, elements = xtp_extract_metadata(file_name)
    header = generate_title(file_name.stem) + header
    s = xtp_table_header(header) if elements is not None else f"{header}\n"
    for elem in elements:
        s += xtp_get_recursive_attributes(elem)

    s += generate_note(file_name.stem)
    return s


def csg_create_rst_table(file_name: Path) -> str:
    """Create an RST table using the metadata in the XML file."""
    header, elements = csg_extract_metadata(file_name)
    s = csg_table_header(header)
    for elem in elements:
        s += csg_get_recursive_attributes(elem)
    return s


def create_parent_folder(path: Path) -> None:
    """Create parent folder for ``path`` if it doesn't exists."""
    if not path.parent.exists():
        path.parent.mkdir(parents=True)


def split_line(line: str, sep: Optional[str] = None) -> str:
    """Split line if larger than ``MAXIMUM_LINE_LENGTH``."""
    if len(line) > MAXIMUM_LINE_LENGTH:
        return wrap_line(line, sep)
    else:
        return line


def wrap_line(line: str, sep: Optional[str]) -> str:
    """Split a line into lines smaller than ``max_len``."""
    acc = [[]]  # type: List[List[str]]
    count = 7  # mulitiple lines start at column number 7
    for word in line.split(sep=sep):
        # If cumulative sum or the length of word is greater than MAXIMUM_LINE_LENGTH
        if any(x > MAXIMUM_LINE_LENGTH for x in (count, len(word))):
            count = 7
            acc.append([])
        # Append words to the last list
        acc[-1].append(word)
        count += 1 + len(word)

    it = iter(" ".join(word for word in line) for line in acc)
    return column_multiline(it)


def generate_title(stem: str) -> str:
    """Generate title in rst format using ``file_name``."""
    return f"{stem}\n{'*' * len(stem)}\n"


def generate_note(stem: str) -> str:
    """Generate note specifying path to the xml file."""
    note = f"""
.. note::
   An *xml* file containing the defaults for the `{stem}` calculator can be created via `-p {stem} -o FILENAME` command line options `
"""
    return note


def multiline(line: str) -> str:
    """Split the comma separated words into lines."""
    xs = line.split(',')
    if len(xs) >= 2:
        return column_multiline(iter(xs))
    else:
        return line


def column_multiline(xs: Iterable[str]) -> str:
    """Create a column with multiple lines."""
    spaces = f"\n{' ':^6s} | "
    return "| " + next(xs) + spaces + f"{spaces}".join(xs)


if __name__ == "__main__":
    main()



"""Visualization module."""
from ase import Atoms
import numpy as np
import matplotlib.pyplot as plt
from .utils import H2EV


class Visualization:
    def __init__(self, atoms: Atoms, save_figure: bool = False):
        self.atoms = atoms
        self.save_figure = save_figure

    def plot_qp_corrections(self):
        qp_corrections = self.atoms._calc.get_qp_corrections()
        qpmin = self.atoms._calc.qpmin
        qpmax = self.atoms._calc.qpmax + 1
        corrected_ks = self.atoms._calc.ks_energies[qpmin:qpmax]
        plt.plot(corrected_ks, qp_corrections, 'ro')
        plt.xlabel('KS Energy (eV)')
        plt.ylabel('QP correction (eV)')
        if self.save_figure:
            plt.savefig("qpcorrections.png")
        else:
            plt.show()

    def plot_absorption_gaussian(
            self, dynamic: bool = False, energy_min: float = 0.0, energy_max: float = 10.0, points: int = 1000,
            sigma: float = 0.2):

        energy, osc = self.atoms._calc.get_oscillator_strengths(dynamic)

        # convert energies from Hartree to eV
        energy *= H2EV  # to eV
        # make a stick plot with oscillator strength
        plt.stem(energy, osc, basefmt=" ")
        # apply unormalized Gaussian lineshape
        e = np.linspace(energy_min, energy_max, points)
        spectrum = 0

        for i, energy_val in enumerate(energy):
            spectrum += osc[i] * self.gaussian(e, energy_val, sigma)

        plt.plot(e, spectrum, 'k', linewidth=2)
        plt.ylim(bottom=0)
        plt.title(f'Gaussian lineshape with sigma = {sigma}eV')
        plt.xlabel('Energy (eV)')
        plt.ylabel('Absorption (arb. units)')
        if self.save_figure:
            plt.savefig("absorption_gaussian.png")
        else:
            plt.show()


    def gaussian(self, x, mu, sig):
        """Non-normalized Gaussian distribution."""
        return np.exp(-0.5 * ((x - mu) / sig) ** 2)

    def plot_dos_gaussian(self, ks=True, qp=True, energy_min: float = -25.0, energy_max: float = 10.0, points: int = 1000,
                          sigma: float = 0.2):

        plt.xlabel('Energy (eV)')
        plt.ylabel('DOS (arb. units)')
        e = np.linspace(energy_min, energy_max, points)
        dos_ks = 0
        if ks:
            ks_energies = self.atoms._calc.ks_energies * H2EV  # to eV
            qpmin = self.atoms._calc.qpmin
            qpmax = self.atoms._calc.qpmax + 1
            for ks_energy in ks_energies[qpmin:qpmax]:
                dos_ks += self.gaussian(e, ks_energy, sigma)
            plt.plot(e, dos_ks, 'k', linewidth=2, label='KS')
        dos_qp = 0
        if qp:
            qp_energies = self.atoms._calc.qp_energies * H2EV
            for qp_energy in qp_energies:
                dos_qp += self.gaussian(e, qp_energy, sigma)
            plt.plot(e, dos_qp, 'r', linewidth=2, label='QP')
        plt.legend()
        if self.save_figure:
            plt.savefig("dos_gaussian.png")
        else:
            plt.show()



#!/usr/bin/env python3
"""Example to compute energies using XTP."""
from pyxtp import xtp, Visualization
from ase import Atoms

def run_energy(save_figure: bool = False):
    """Run energy workflow."""
    # define a molecule
    atoms = Atoms('CO', positions=([0,0,0],[1.4,0,0]))

    # define the calculator
    calc = xtp(nthreads=2)

    # change basis sets to a smaller one
    calc.options.dftpackage.basisset = 'def2-svp'
    calc.options.dftpackage.auxbasisset = 'aux-def2-svp'
    calc.options.logging_file = 'CO_energy.log'

    # attach the calculator
    atoms.calc = calc

    # run the calculations
    atoms.get_potential_energy()

    # only needed, if no run was performed but an existing HDF5 is read
    # atoms.read('pyvotca/examples/example.orb')

    # Getting the plotting functions
    viz = Visualization(atoms, save_figure=save_figure)
    # plotting QP corrections
    # viz.plot_qp_corrections()
    # plotting absorption spectrum
    viz.plot_absorption_gaussian()


if __name__ == "__main__":
    run_energy()



#!/usr/bin/env python3
"""Examples to show xtp_binds usage."""
from pyxtp_binds import xtp_binds
from pathlib import Path
from multiprocessing import cpu_count

def run_mapchecker(nThreads: int, path_examples: Path) -> None:
    """Run the mapchecker calculator."""
    # mapchecker
    print("Calling mapchecker")
    # Look for the xml file to use as input
    xml_file = (path_examples / "mapchecker.xml").absolute().as_posix()
    state_file = (path_examples / "state.hdf5").absolute().as_posix()
    # Input to call the calculator
    dict_inp = {
        "xmlFile": xml_file,
        "stateFile": state_file,
        "nThreads": str(nThreads),
        "nFrames": '1',
        "firstFrame": '0',
        "save": '1'  # xtp_binds casts it to bool False
    }
    xtp_binds.call_calculator("mapchecker", dict_inp)

    # Check that the files exists
    expected = {"md_segments_step_0.pdb", "mp_segments_e_step_0.pdb",
                "mp_segments_h_step_0.pdb", "qm_segments_n_step_0.pdb"}
    pdb_files = set(p.as_posix() for p in Path(".").glob("*_step_0.pdb"))
    if not all(x in expected for x in pdb_files):
        raise AssertionError("map checker failed")


def run_examples():
    """Call the xtp_binds interface."""
    nThreads = cpu_count()
    path_examples = Path("files_examples")

    run_mapchecker(nThreads, path_examples)

if __name__ == "__main__":
    run_examples()



# -*- coding: utf-8 -*-

import logging

from .__version__ import __version__
from .xtp import xtp
from .visualization import Visualization

logging.getLogger(__name__).addHandler(logging.NullHandler())

__author__ = "Bjoern Baumeier"

__all__ = ['xtp','Visualization']



"""Module defining the Option class."""
import os
from pathlib import Path
from types import new_class
from typing import Callable

from lxml import etree
import xmltodict

from pyxtp.xml_editor import (
    remove_empty_text_elements,
    insert_linked_xmlfiles,
    update_xml_text,
)


class Options:
    # NOTE: if you add a class/instance attribute, please update __slots__
    __slots__ = ("_xml_tree", "_opts")

    def __init__(self, xml_file: str | Path, xml_attribs=True, set_default=False):
        """Create options object from the input xml files"""

        # parse xmlfile and replace links to the embedded xml files
        xml_file = Path(xml_file)
        self._xml_tree = insert_linked_xmlfiles(
            etree.parse(xml_file), base_path=f"{xml_file.parent}"
        )
        dict_data = xmltodict.parse(
            etree.tostring(self._xml_tree), xml_attribs=xml_attribs
        )
        name = xml_file.stem
        self._opts = getattr(make_options(name, dict_data, set_default).options, name)

    def __getattr__(self, name: str):
        if name in self.__slots__:
            return getattr(self, name)
        return getattr(self._opts, name)

    def __setattr__(self, name: str, value):
        if name in self.__slots__:
            super().__setattr__(name, value)
        else:
            setattr(self._opts, name, value)

    def __dir__(self):
        from_opts = [attr for attr in dir(self._opts) if not attr.startswith("_")]
        return [*self.__slots__, *from_opts]

    def _write_xml(self, filename: str) -> None:
        """update and clean the xml and then write the input file

        Args:
            filename (str): filename in the
        """

        # create a dict of the user providedoptions
        options = self._opts.to_flat_dict()

        # update the options from the user provided input
        xml_tree_cpy = update_xml_text(self._xml_tree, options)

        # remove the empty elements
        xml_tree_cpy = remove_empty_text_elements(xml_tree_cpy)

        # write the file
        xml_tree_cpy.write(filename)


class _Opts_t:
    __slots__ = ()
    def to_dict(self) -> dict:
        """Convert a config object to a dictionary, skip keys that are not set"""
        res = {}
        for attr in dir(self):
            if attr.startswith("_"):
                continue
            value = getattr(self, attr)
            if value is None or isinstance(value, Callable):
                continue
            if isinstance(value, _Opts_t):  # recurse
                if _val := value.to_dict():  # drop empty
                    res[attr] = _val
            else:  # leaf node
                res[attr] = value
        return res

    def to_flat_dict(self) -> dict:
        """Transform options object into a flat dictionary.

        Returns:
            dict: dictionary maps the namespace
        """

        def _flatten_dict(data: dict, path: str = "") -> dict:
            """Flatten a nested dictionary.

            Args:
                data (dict): input dictionary
                path (str, optional): basepath in the structure. Defaults to ''.

            Returns:
                dict: dictionary maps the namespace
            """
            output = dict()
            for key, value in data.items():
                if isinstance(value, dict):
                    output.update(_flatten_dict(value, path=os.path.join(path, key)))
                else:
                    if not key.startswith("_"):
                        output[os.path.join(path, key)] = value
            return output

        return _flatten_dict(self.to_dict())


def make_options(name: str, xml_dict: dict, set_default: bool = True):
    """Create a config object out of an XML dictionary

    Parameters
    ----------
    name : str
      Name for options type

    xml_dict : dict
      Dictionary generated from the XML configuration files

    set_default : bool (default: True)
      Whether to set default values in the config object

    Returns
    -------
    name_t

    """
    # detect leaf, return early w/ default
    children = [k for k in xml_dict if not k.startswith("@")]
    if len(children) == 0:  # leaf node
        default = xml_dict.get("@default", None)
        if default == "OPTIONAL":
            return None
        if not set_default:
            return ''
        return default

    # create type
    attrs, slots = {}, {}
    for key, value in xml_dict.items():
        if isinstance(value, dict):
            doc = [v for k, v in value.items() if k.startswith("@")]
            slots[key] = "\n".join(doc)
            attrs[key] = make_options(key, value, set_default)
    cls = new_class(
        f"{name}_t",
        bases=(_Opts_t,),
        exec_body=lambda ns: ns.update({"__slots__": slots}),
    )
    obj = cls()

    # assign slots & defaults
    for k, v in attrs.items():  # no @keys here
        if isinstance(v, _Opts_t) or set_default:
            setattr(obj, k, v)
        elif not set_default:
            setattr(obj, k, '')
        else:
            raise ValueError('Something went wrong')
    return obj



import os
import numpy as np
from typing import Dict, Union, List , Optional, Tuple
import h5py
from pathlib import Path

from ase import Atoms
from ase.calculators.calculator import Calculator, equal, PropertyNotImplementedError
from .capture_standard_output import capture_standard_output
from pyxtp_binds import xtp_binds
from .options import Options
from .utils import BOHR2ANG

Pathlike = Union[Path, str]


def numeric_force_xtp(atoms, a, i, d=0.001, opt_forces: dict = None):
    """Compute numeric force on atom with index a, Cartesian component i,
    with finite step of size d, and property name (with level level when applicable)
    """
    if opt_forces is None:
        name, level, dynamic = 'energy', 0, False
    else:
        name, level, dynamic = opt_forces['energy'], opt_forces['level'], opt_forces['dynamic']

    p0 = atoms.get_positions()
    p = p0.copy()
    p[a, i] += d
    atoms.set_positions(p, apply_constraint=False)
    eplus = atoms._calc.get_total_energy(name=name, level=level, dynamic=dynamic)

    atoms._calc.reset_results()

    p[a, i] -= 2 * d
    atoms.set_positions(p, apply_constraint=False)
    eminus = atoms._calc.get_total_energy(name=name, level=level, dynamic=dynamic)
    atoms.set_positions(p0, apply_constraint=False)
    return (eminus - eplus) / (2 * d)

class xtp(Calculator):
    """This the ASE-calculator for doing XTP calculation."""

    implemented_properties = ['energy', 'forces', 'singlets',
                              'triplets', 'qp', 'ks',
                              'qp_pert', 'oscillator_strength']

    def __init__(self,
                 restart=None,
                 *,
                 label=None,
                 atoms=None,
                 nthreads=1,
                 options=None,
                 directory='./',
                 **kwargs):

        Calculator.__init__(self, restart=restart, label=label, atoms=atoms,
                            directory=directory, **kwargs)

        if options is None:
            self.set_default_options()
        else:
            self.set_from_options(options)

        self.set(**kwargs)
        self.select_force()

        self.nthreads = nthreads

        self.has_data = False
        self.has_forces = False

        self.hdf5_filename = None
        self.logfile = 'dftgwbse.log'

    def set_default_options(self) -> Options:
        """Creates a default option object

        Raises:
            RuntimeError: if VOTCASHARE is not defined

        Returns:
            Options: _description_
        """
        votcashare = os.environ.get("VOTCASHARE")
        if votcashare is None:
            msg = (
                "pyxtp: cannot find Votca installation, "
                "please set the VOTCASHARE environment variable"
            )
            raise RuntimeError(msg)
        file_name = f"{votcashare}/xtp/xml/dftgwbse.xml"

        self.options = Options(file_name)
        self.parameters = self.options.to_flat_dict()

    def set(self, **kwargs) -> Dict:
        """Set parameters like set(key1=value1, key2=value2, ...).

        A dictionary containing the parameters that have been changed
        is returned.
        """

        translation_dic = {'xc': 'dftpackage/functional',
                           'basis': 'dftpackage/basisset',
                           'charge': 'dftpackage/charge'}

        changed_parameters = {}

        for key, value in kwargs.items():

            if key in translation_dic:
                key = translation_dic[key]

            if key in self.parameters:
                oldvalue = self.parameters.get(key)
                if key not in self.parameters or not equal(value, oldvalue):
                    changed_parameters[key] = value
                    self.parameters[key] = value

                    split_key = key.split('/')
                    element = self.options
                    for k in split_key[:-1]:
                        element = element.__getattr__(k)
                    element.__setattr__(split_key[-1], value)
            else:
                print(f'Option {key} not available in xtp')

        if self.discard_results_on_any_change and changed_parameters:
            self.reset()

        return changed_parameters

    def set_from_options(self, options: Options) -> None:
        """Set the options from an instance of XTPOptions

        Args:
            options (XTPOptions): user defined options
        """
        self.options = options
        opt_dict = options.to_flat_dict()

        if not self.parameters:
            self.parameters = opt_dict

        self.set(**opt_dict)


    def set_atoms(self, atoms: Atoms):
        """Set atomic positions

        Args:
            atoms (Atoms): atoms in ASE format
        """
        self.atoms = atoms

    def reset(self):
        """Clear all information from old calculation."""

        self.atoms = None
        self.reset_results()

    def reset_results(self):
        """Clear all the results of previous calculations but keep atoms data
        """
        self.results = {}
        self.select_force()
        self.has_forces = False
        self.has_data = False


    def calculate(self, atoms: Atoms = None,
                  properties: List[str] = ['energy'],
                  system_change : List[str] = None) -> None:
        """Calculate things

        Args:
            atoms (Atoms, optional): atoms in ase format. Defaults to None.
            name (str, optional): basename for the files. Defaults to None.
        """
        if any([p in properties for p in ['energy',
                                          'singlets', 'triplets',
                                          'ks', 'qp', 'qp_pert',
                                          'transition_dipoles']]):
            self.calculate_energies(atoms=atoms)

        if 'forces' in properties:
            self.calculate_numerical_forces('energy')

    def calculate_energies(self, atoms: Atoms = None) -> None:
        """Compute all the energies of the dft+gw+bse pipeline

        Args:
            atoms (Atoms, optional): atoms in ase format. Defaults to None.
        """

        Calculator.calculate(self, atoms)
        atoms = self.atoms

        # write input files
        if self.label is None:
            xyzname = self.atoms.get_chemical_formula()
        else:
            xyzname = self.label

        xyz_filename = xyzname + '.xyz'
        input_filename = 'dftgwbse.xml'
        self.atoms.write(xyz_filename)
        self.options.job_name = xyzname
        self.options._write_xml(input_filename)

        # Runs VOTCA and moves results a job folder, if requested
        if not Path(self.directory).exists():
            os.makedirs(self.directory)

        # current dftgwbse file
        path_dftgwbse = (Path(input_filename)).absolute().as_posix()

        # Call the tool and capture the standard output
        output = capture_standard_output(
            xtp_binds.call_tool, "dftgwbse", self.nthreads,  path_dftgwbse)

        with open(xyzname + ".out", "w") as handler:
            handler.write(output)

        # copy orbfile, in directory
        orbname = f'{xyzname}.orb'
        self.hdf5_filename = os.path.join(self.directory, orbname)
        os.replace(f'{xyzname}.orb', self.hdf5_filename)

        # Reads energies from an existing HDF5 orb file
        self.read_hdf5_data(self.hdf5_filename)

    @classmethod
    def read_atoms(cls, filename: Pathlike, label = None, **kwargs) -> Atoms:
        """Read an exisiting hdf5 file containing results from a previous run
        Args:
            filename (Pathlike): name of the file
        """
        if os.path.splitext(filename)[-1] != '.orb':
            raise ValueError('The file should be a .orb file generated by VOTCA-XTP')
        return cls(label=label, **kwargs).load_atoms(filename)

    def load_atoms(self, filename: Pathlike) -> Atoms:
        """Load the atoms data from a hdf5 file

        Args:
            filename (Pathlike): filename

        Returns:
            Atoms: instance of atoms
        """
        self.read_hdf5_data(filename)
        atoms = self.atoms.copy()
        atoms.calc = self
        return atoms

    def read_hdf5_data(self, hdf5_filename: Pathlike) -> None:
        """Read data from the orb (HDF5) file

        Args:
            hdf5_filename (Pathlike): name for the hdf5 file containing the data
        """
        with h5py.File(hdf5_filename, 'r') as handler:
            orb = handler['QMdata']
            # get coordinates
            atoms = orb['qmmolecule']['qmatoms']
            # coordinates are stored in Bohr!
            arr = [(atom['element'][0].decode(), BOHR2ANG * np.array(
                [atom['posX'][0], atom['posY'][0], atom['posZ'][0]], dtype=float)) for atom in atoms]
            elements_in, coordinates_in = tuple(zip(*arr))

            if self.atoms is None:
                self.atoms = Atoms(elements_in, positions=coordinates_in)
            else:
                self.check_molecule_integrity(elements_in, coordinates_in)

            self.has_xyz = True

            self.homo = int(orb.attrs['occupied_levels'][0]) - 1
            self.DFTenergy = float(orb.attrs['qm_energy'][0])
            self.qp_energies = read_flatten_array(orb, 'QPpert_energies')
            self.qp_energies_diag, self.ks_energies, self.bse_singlet_energies, self.bse_triplet_energies = [
                read_flatten_array(orb, x, 'eigenvalues') for x in ('QPdiag', 'mos', 'BSE_singlet', 'BSE_triplet')]

            self.bse_singlet_energies_dynamic, self.bse_triplet_energies_dynamic = [
                read_flatten_array(orb, f"BSE_{x}_dynamic") for x in ("singlet", "triplet")]
            self.qpmin = int(orb.attrs['qpmin'][0])
            self.qpmax = int(orb.attrs['qpmax'][0])
            td = orb['transition_dipoles']
            self.transition_dipoles = np.array(
                [td[dset][()] for dset in td.keys()])

            self.has_data = True

        # create the results
        self.results = {
            'energy': self.DFTenergy,
            'singlets': self.bse_singlet_energies,
            'triplets': self.bse_triplet_energies,
            'ks': self.ks_energies,
            'qp': self.qp_energies_diag,
            'transition_dipoles': self.transition_dipoles,
            'qp_pert': self.qp_energies,
        }


    def get_total_energy(self, name: str = 'energy', level: int = 0, dynamic: bool = False,
                         atoms = None, allow_calculation=True) -> float:
        """Wrap call to individual total energy functions.

        Args:
            name (str): name of the energy required
            level (int): if multiple level in the energy required, specify which one
            dynamic (bool, optional): dynamic propoerty. Defaults to False.
            atoms (Atoms, optional): atoms to compute the properties over (None)
            allow_calculation(bool, optional): allow calculation if True (True)

        Raises:
            Exception: if kind not recognized

        Returns:
            float: value of the energy (level) required
        """

        if name not in self.implemented_properties:
            raise PropertyNotImplementedError('{} property not implemented'
                                              .format(name))
        if name == "forces":
            raise PropertyNotImplementedError('{} property must be calculated with \
                .get_forces()'.format(name))

        if name == "oscillator_strength":
            raise PropertyNotImplementedError('{} property must be calculated with \
                .get_oscillator_strength()'.format(name))

        if atoms is None:
            atoms = self.atoms
            system_changes = []
        else:
            system_changes = self.check_state(atoms)
            if system_changes:
                self.reset()

        if name not in self.results:
            if not allow_calculation:
                return None
            self.calculate(properties=[name])

        if name  == 'energy':
            return self.get_dft_energy()
        elif name == 'ks':
            return self.get_ks_total_energy(level)
        elif name == 'qp_pert':
            return self.get_qp_total_energy(level)
        elif name == 'qp':
            return self.get_qp_total_energy(level)
        elif name == 'singlets' and not dynamic:
            return self.get_bse_singlet_total_energy(level)
        elif name == 'singlets' and dynamic:
            return self.get_bse_singlet_dynamic_total_energy(level)
        elif name == 'triplets' and not dynamic:
            return self.get_bse_triplet_total_energy(level)
        elif name == 'triplets' and dynamic:
            return self.get_bse_triplet_dynamic_total_energy(level)
        else:
            raise Exception(
                f'Energy of kind {name} is not available!')

    def get_dft_energy(self) -> float:
        """Return the DFT total energy.

        Returns:
            float: dft total energy
        """
        self.check_data()
        return self.DFTenergy

    def get_ks_total_energy(self, level='') -> float:
        """Return the excited state KS total energy.

        Args:
            level (str, optional): _description_. Defaults to ''.

        Returns:
            float: Kohn Sham total energy
        """
        self.check_data()

        lumo = self.homo + 1

        total_energy = self.DFTenergy
        if (level < lumo):
            return(total_energy - self.ks_energies[level])
        elif level < len(self.ks_energies):
            return(total_energy + self.ks_energies[level])
        else:
            print("Requested KS level {} does not exist.")
            return 0.0

    def get_qp_total_energy(self, level='') -> float:
        """Return the excited state QP total energy.

        Args:
            level (str, optional): _description_. Defaults to ''.

        Returns:
            float: Quasi Particle total energy
        """

        self.check_data()

        lumo = self.homo + 1

        total_energy = self.DFTenergy
        if (level < lumo):
            return(total_energy - self.qp_energies[level - self.qpmin])
        elif level < len(self.ks_energies):
            return(total_energy + self.qp_energies[level - self.qpmin])
        else:
            print("Requested QP level {} does not exist.")
            return 0.0

    def get_qp_diag_total_energy(self, level='') -> float:
        """Return the excited state diag QP total energy.

        Args:
            level (str, optional): _description_. Defaults to ''.

        Returns:
            float: Quasi Particle total energy
        """
        self.check_data()

        lumo = self.homo + 1

        total_energy = self.DFTenergy
        if (level < lumo):
            return(total_energy - self.qp_energies_diag[level - self.qpmin])
        elif level < len(self.ks_energies):
            return(total_energy + self.qp_energies_diag[level - self.qpmin])
        else:
            print(f"Requested diag QP {level} does not exist.")
            return 0.0

    def get_bse_singlet_total_energy(self, level: int) -> float:
        """Return the excited state BSE Singlet total energy.

        Args:
            level (int): singlet level required.

        Returns:
            float: Singlet Energy required
        """
        msg = f"Requested BSE singlet {level} does not exist."
        return self.check_and_read(level, "bse_singlet_energies", msg)

    def get_bse_triplet_total_energy(self, level: int) -> float:
        """Return the excited state BSE Triplet total energy.

        Args:
            level (int): triplet level required.

        Returns:
            float: Triplet Energy required
        """
        msg = f"Requested BSE triplet {level} does not exist."
        return self.check_and_read(level, "bse_triplet_energies", msg)

    def get_bse_singlet_dynamic_total_energy(self, level: int) -> float:
        """Return the excited state BSE Singlet dynamic energy.

        Args:
            level (int): singlet level required.

        Returns:
            float: Singlet Energy required
        """
        msg = f"Requested dynamic BSE singlet {level} does not exist."
        return self.check_and_read(level, "bse_singlet_energies_dynamic", msg)

    def get_bse_triplet_dynamic_total_energy(self, level: int) -> float:
        """Return the excited state BSE Triplet dynamic energy.

        Args:
            level (int): triplet level required.

        Returns:
            float: Triplet Energy required
        """
        msg = f"Requested dynamic BSE triplet level {level} does not exist."
        return self.check_and_read(level, "bse_triplet_energies_dynamic", msg)

    def get_qp_corrections(self) -> np.ndarray:
        """Return the quasi particle correction energies

        Returns:
            np.ndarray: correction energies
        """

        self.check_data()

        qp_corrections = self.qp_energies -\
            self.ks_energies[self.qpmin:self.qpmin + len(self.qp_energies)]

        return qp_corrections.flatten()

    def get_oscillator_strengths(self, dynamic: bool = False) -> Tuple[np.ndarray, np.ndarray]:
        """Retrieve oscillator strenghts' values.

        Args:
            dynamic (bool, optional): Return dynamic values. Defaults to False.

        Returns:
            Tuple[np.ndarray, np.ndarray]: energy, oscillator strength
        """
        self.check_data()

        # get energies/oscillator strengths
        if dynamic:
            energy = self.bse_singlet_energies_dynamic
        else:
            energy = self.bse_singlet_energies
        osc = [(2. / 3.) * e * (t ** 2).sum()
               for e, t in zip(energy, self.transition_dipoles)]

        return energy, np.array(osc)

    def select_force(self, energy: str = 'energy', level: int = 0, dynamic: bool = False):
        """Set up which energy term and energy level we want to compute the forces on

        Args:
            energy (str, optional): _description_. Defaults to 'energy'.
            energy_level (int, optional): _description_. Defaults to 0.
        """
        self.option_forces = {'energy': energy, 'level': level, 'dynamic': dynamic}

    def calculate_numerical_forces(self, eps = 0.001, atoms: Atoms = None) -> np.ndarray:
        """Retreive the atomic forces

        Args:
            kind (str): kind of particle/excitation (choices: BSE_singlet, BSE_triplet, QPdiag, QPpert and dft_tot)
                        and optionally the energy level if not provided all energy levels will be returned
            energy_level (int): level of the required energy

        Returns:
            np.ndarray: atomic gradients
        """
        if atoms is None:
            atoms = self.atoms


        forces = []
        for a in range(len(atoms)):
            _force = []
            for i in range(3):
                new_atoms = Atoms(atoms.get_chemical_symbols(), positions=atoms.get_positions())
                new_atoms.calc = xtp(options=self.options, nthreads=self.nthreads)
                _force.append(numeric_force_xtp(new_atoms, a, i, eps, self.option_forces))
            forces.append(_force)
        self.results['forces'] = np.array(forces) #* Hartree / Bohr ?
        self.has_forces = True
        return self.results['forces']

    def get_forces(self, atoms: Atoms = None, eps = 0.001) -> np.ndarray:
        """get the forces or calculate them if they are not present

        Args:
            kind (str, optional): _description_. Defaults to 'energy'.
            energy_level (int, optional): _description_. Defaults to 0.
            eps (float, optional): _description_. Defaults to 0.001.

        Returns:
            np.ndarray: atomic forces
        """
        if self.check_forces():
            return self.results['forces']
        else:
            return self.calculate_numerical_forces(atoms=atoms)

    def read_forces_from_logfile(self, logfile: Pathlike) -> None:
        """Read Forces from VOTCA logfile

        Args:
            logfile (Pathlike): name of the logfile
        """
        fil = open(logfile, 'r', encoding='utf-8')
        lines = fil.readlines()
        fil.close()
        getgrad = "no"
        for line in lines:
            if line.find('ENGRAD') >= 0:
                getgrad = "yes"
                gradients = []
                tempgrad = []
                continue
            if getgrad == "yes" and "Saving" not in line:
                if "Total" not in line:
                    grad = line.split()
                    tempgrad.append(float(grad[3]))
                    tempgrad.append(float(grad[4]))
                    tempgrad.append(float(grad[5]))
                    gradients.append(tempgrad)
                    tempgrad = []
            if 'Saving' in line:
                getgrad = "no"
        self.atomic_forces = -np.array(gradients) #* Hartree / Bohr

    def check_forces(self):
        """Check if forces are really present.
           The MD and optimization of ASE sometimes remove results but do not
           reset has_forces
        """
        if (self.has_forces) and ('forces' in self.results):
            return True
        self.has_forces = False
        return False


    def check_data(self):
        """Check that there is data in the molecule."""
        if not self.has_data:
            raise Exception("No energy has been stored!")

    def check_and_read(self, level: int, prop: str, msg: str) -> float:
        """Check that there is data available and retrieve it."""
        self.check_data()
        if level < len(getattr(self, prop)):
            return(self.DFTenergy + getattr(self, prop)[level])
        else:
            print(msg)

    def check_molecule_integrity(self, other_elements: List[str], other_coordinates: List[np.ndarray]):
        """Compare the atoms from self with the one stored in the HDF5."""
        for k, (elem, coord, other_elem, other_coord) in enumerate(
                zip(self.atoms.get_chemical_symbols(), self.atoms.get_positions(),
                    other_elements, other_coordinates)):
            if elem != other_elem:
                raise Exception(
                    f'Element {elem} with index {k} in molecule differs from element {other_elem} in orb file!')

            if not np.allclose(coord, other_coord):
                raise Warning(
                    f'Molecular coordinates of element {k} {coord} differ from coordinates in orb file {other_coord}')

def read_flatten_array(group: h5py.Group, key1: str, key2: Optional[str] = None):
    """Read an array from h5py handler and flatten it."""
    if key2 is None:
        arr = group[key1][()]
    else:
        arr = group[key1][key2][()]

    return arr.flatten()







#!/usr/bin/env python3
"""Example to perform a gradient calculation."""
from pyxtp import xtp
from ase import Atoms
import numpy as np

def run_gradient() -> np.ndarray:
    """Compute gradient."""
    # define a molecule
    atoms = Atoms('CO', positions=([0,0,0],[1.4,0,0]))

    # define the calculator
    calc = xtp(nthreads=2)
    calc.select_force(energy='singlets', level=0, dynamic=False)

    # this allows to change all options
    # calc.options.dftpackage.functional = 'PBE'
    calc.options.dftpackage.basisset = 'def2-svp'
    calc.options.dftpackage.auxbasisset = 'aux-def2-svp'

    # attach the calculator
    atoms.calc = calc

    # run for the molecule in its geometry
    return atoms.get_forces()

if __name__ == "__main__":
    run_gradient()



#!/usr/bin/env python3
"""Example to optimize the geometry of CO"""
from pyxtp import xtp
from ase.io import write
from ase.build import molecule
from ase.optimize import QuasiNewton

atoms = molecule('CO')
atoms.rattle()

calc = xtp(nthreads=2)
calc.select_force(energy='singlets', level=0, dynamic=False)
atoms.calc = calc

atoms.get_forces()

dyn = QuasiNewton(atoms, trajectory='test.traj')
dyn.run(fmax=0.01)
write('final.xyz', atoms)



from functools import reduce
import os, os.path

import pytest
from lxml import etree
import xmltodict

from pyxtp.options import Options
from pyxtp.xml_editor import insert_linked_xmlfiles


@pytest.fixture(scope="module")
def dftgwbse_xml():
    votcashare = os.environ.get("VOTCASHARE")
    if votcashare is None:
        msg = (
            "pyxtp: cannot find Votca installation, "
            "please set the VOTCASHARE environment variable"
        )
        raise RuntimeError(msg)
    return f"{votcashare}/xtp/xml/dftgwbse.xml"


class TestOptions:
    @pytest.fixture
    def opts(self, dftgwbse_xml):
        return Options(dftgwbse_xml, set_default=True)

    @pytest.fixture
    def xml_dict(self, dftgwbse_xml):
        path = dftgwbse_xml
        parent = os.path.dirname(path)
        xml_tree = insert_linked_xmlfiles(etree.parse(path), base_path=parent)
        return xmltodict.parse(etree.tostring(xml_tree), xml_attribs=True)

    def test_completion(self, opts, xml_dict):
        # check tab completion of properties
        expected = set(attr for attr in dir(opts) if not attr.startswith("to_"))
        from_xml = list(
            k for k in xml_dict["options"]["dftgwbse"] if not k.startswith("@")
        )
        if expected != {*opts.__slots__, *from_xml}:
            raise AssertionError("Error in test_completion")

    @pytest.mark.parametrize(
        "attr, val", [("job_name", "options_test"), ("mpsfile", "somefile")]
    )
    def test_attr_valid(self, opts, attr, val):
        setattr(opts, attr, val)

    def test_attr_invalid(self, opts):
        with pytest.raises(AttributeError, match="jobname"):
            opts.jobname = "options_test"

    @pytest.mark.parametrize(
        "attr, expected", [("job_name", "system"), (("dftpackage", "spin"), "1")]
    )
    def test_attr_default(self, opts, attr, expected):
        if isinstance(attr, str):
            val = getattr(opts, attr)
        else:
            val = reduce(getattr, attr, opts)
        if expected != val:
            raise AssertionError("Error in test_attr_default")